searchData={"items":[{"type":"module","title":"Torus","doc":"Torus is a plug-and-play library that seamlessly integrates PostgreSQL's search into Ecto, streamlining the construction of advanced search queries.","ref":"Torus.html"},{"type":"module","title":"Usage - Torus","doc":"The package can be installed by adding `torus` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:torus, \"~> 0.2.1\"}\n  ]\nend\n```\n\nThen, in any query, you can (for example) add a full-text search:\n\n```elixir\nPost\n|> Torus.full_text_dynamic([p], [p.title, p.body], \"uncovered hogwarts\")\n|> Repo.all()\n```\n\nSee `full_text_dynamic/5` for more details.","ref":"Torus.html#module-usage"},{"type":"module","title":"6 types of search: - Torus","doc":"1. **Pattern matching**: Searches for a specific pattern in a string.\n\n   ```elixir\n   iex> insert_posts!([\"Wand\", \"Magic wand\", \"Owl\"])\n   ...> Post\n   ...> |> Torus.ilike([p], [p.title], \"wan%\")\n   ...> |> select([p], p.title)\n   ...> |> Repo.all()\n   [\"Wand\"]\n   ```\n\n   See `like/5`, `ilike/5`, and `similar_to/5` for more details.\n\n1. **Similarity**: Searches for items that are closely alike based on attributes, often using measures like cosine similarity or Euclidean distance.\n\n   ```elixir\n   iex> insert_posts!([\"Magic wand\", \"Wand\", \"Owl\"])\n   ...> Post\n   ...> |> Torus.similarity([p], [p.title], \"want\", limit: 2)\n   ...> |> select([p], p.title)\n   ...> |> Repo.all()\n   [\"Wand\", \"Magic wand\"]\n   ```\n\n   See `similarity/5` for more details.\n\n1. **Text Search Vectors**: Uses term-document matrix vectors for **full-text search**, enabling\n   efficient querying and ranking based on term frequency. - [PostgreSQL: Documentation: 17: Chapter 12. Full Text Search](https://www.postgresql.org/docs/current/textsearch.html)\n\n   ```elixir\n   iex> insert_posts!([\"Wand\", \"Owl\", \"What an amazing cloak\"])\n   ...> Post\n   ...> |> Torus.full_text_dynamic([p], [p.title], \"what a cloak\")\n   ...> |> select([p], p.title)\n   ...> |> Repo.all()\n   [\"What an amazing cloak\"]\n   ```\n\n   See `full_text_dynamic/5` for more details.\n\n1. **Semantic Search**: Understands the contextual meaning of queries to match and retrieve related content, often utilizing natural language processing.\n   [Semantic Search with PostgreSQL and OpenAI Embeddings](https://towardsdatascience.com/semantic-search-with-postgresql-and-openai-embeddings-4d327236f41f)\n\n   Will be added soon.\n\n1. **Hybrid Search**: Combines multiple search techniques (e.g., keyword and semantic) to leverage their strengths for more accurate results.\n\n   Will be added soon.\n\n1. **3rd Party Engines/Providers**: Utilizes external services or software specifically designed for optimized and scalable search capabilities, such as Elasticsearch or Algolia.","ref":"Torus.html#module-6-types-of-search"},{"type":"module","title":"Torus support - Torus","doc":"For now, Torus supports pattern match, similarity, and full-text search, with plans to expand support further. These docs will be updated with more examples on which search type to choose and how to make them more performant (by adding indexes or using specific functions).","ref":"Torus.html#module-torus-support"},{"type":"module","title":"Debugging your queries - Torus","doc":"Torus offers a few helpers to debug, explain, and analyze your queries before using them on production. See `Torus.QueryInspector` for more details.","ref":"Torus.html#module-debugging-your-queries"},{"type":"function","title":"Torus.apply_case/3","doc":"","ref":"Torus.html#apply_case/3"},{"type":"macro","title":"Torus.full_text_dynamic/5","doc":"Full text search with rank ordering. Accepts a list of columns to search in.\nCleans the term, so it can be input directly by the user. The default preset of\nsettings is optimal for most cases.","ref":"Torus.html#full_text_dynamic/5"},{"type":"macro","title":"Options - Torus.full_text_dynamic/5","doc":"* `:language` - language used for the search. Defaults to `\"english\"`.\n  * `:prefix_search` - whether to apply prefix search.\n    - `true` (default) - the term is treated as a prefix\n    - `false` - only counts full-word matches\n  * `:term_function` - function used to convert the term to `ts_query`. Can be one of:\n    - `:websearch_to_tsquery` (default) - converts term to a tsquery, normalizing\n    words according to the specified or default configuration. Quoted word sequences\n    are converted to phrase tests. The word “or” is understood as producing an OR\n    operator, and a dash produces a NOT operator; other punctuation is ignored. This\n    approximates the behavior of some common web search tools.\n    - `:plainto_tsquery` - converts term to a tsquery, normalizing words according\n    to the specified or default configuration. Any punctuation in the string is\n    ignored (it does not determine query operators). The resulting query matches\n    documents containing all non-stopwords in the term.\n    - `:phraseto_tsquery` - converts term to a tsquery, normalizing words according\n    to the specified or default configuration. Any punctuation in the string is\n    ignored (it does not determine query operators). The resulting query matches\n    phrases containing all non-stopwords in the text.\n  * `:rank_function` - function used to rank the results.\n    - `:ts_rank_cd` (default) - computes a score showing how well the vector matches\n    the query, using a cover density algorithm. See [Ranking Search Results](https://postgresql.org/docs/current/interactive/textsearch-controls.html#TEXTSEARCH-RANKING) for more details.\n    - `:ts_rank` - computes a score showing how well the vector matches the query.\n  * `:rank_weights` - a list of weights for each column. Defaults to `[:A, :B, :C, :D]`.\n  The length of weights (if provided) should be the same as the length of the columns we search for.\n  A single weight can be either a string or an atom. Possible values are:\n    - `:A` - 1.0\n    - `:B` - 0.4\n    - `:C` - 0.2\n    - `:D` - 0.1\n  * `:rank_normalization` - a string that specifies whether and how a document's\n  length should impact its rank. The integer option controls several behaviors, so\n  it is a bit mask: you can specify one or more behaviors using `|` (for example, `2|4`).\n    - `0` (default for `ts_rank`) - ignores the document length\n    - `1`  - divides the rank by 1 + the logarithm of the document length\n    - `2`  - divides the rank by the document length\n    - `4` (default for `ts_rank_cs`)  - divides the rank by the mean harmonic\n    distance between extents (this is implemented only by `ts_rank_cd`)\n    - `8`  - divides the rank by the number of unique words in document\n    - `16` -  divides the rank by 1 + the logarithm of the number of unique words in\n    document\n    - `32` - divides the rank by itself + 1\n  * `:order` - describes the ordering of the results. Possible values are\n    - `:desc` (default) - orders the results by similarity rank in descending order.\n    - `:asc` - orders the results by similarity rank in ascending order.\n    - `:none` - doesn't apply ordering at all.\n  * `:filter_type`\n    - `:or` (default) - uses `OR` operator to combine different column matches.\n    Selecting this option means that the search term won't match across columns.\n    - `:concat` - joins the columns into a single tsvector and searches for the\n    term in the concatenated string containing all columns.\n    - `:none` - doesn't apply any filtering and returns all results.\n  * `:nullable_columns` - a list of columns which values can take `null`s, so that\n  they can be escaped during the search. Defaults to all qualifiers. When passed -\n  speeds up the search. For example `[p.title, a.first_name]` would mean that only\n  `p.title` and `a.first_name` columns can take `null`s, whereas other columns are\n  guaranteed by the caller to be non-nullable.","ref":"Torus.html#full_text_dynamic/5-options"},{"type":"macro","title":"Example usage - Torus.full_text_dynamic/5","doc":"iex> insert_post!(title: \"Hogwarts Shocker\", body: \"A spell disrupts the Quidditch Cup.\")\n    ...> insert_post!(title: \"Diagon Bombshell\", body: \"Secrets uncovered in the heart of Hogwarts.\")\n    ...> insert_post!(title: \"Completely unrelated\", body: \"No magic here!\")\n    ...>  Post\n    ...> |> Torus.full_text_dynamic([p], [p.title, p.body], \"uncov hogwar\", filter_type: :concat)\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Diagon Bombshell\"]","ref":"Torus.html#full_text_dynamic/5-example-usage"},{"type":"macro","title":"Optimizations - Torus.full_text_dynamic/5","doc":"- Store precomputed tsvector in a separate column, add a GIN index to it, and use\n  `full_text_stored/5`. See more on how to add an index and how to store a column in\n  the `full_text_stored/5` docs. If that's not feasible, read on.\n\n  - Pass implicitly `nullable_columns` so that we're leveraging existing non-coalesce\n  GIN indexes (if present) and are not doing unneeded work.\n\n  - Add a GIN ts_vector index on the column(s) you search in.\n  Use `Torus.QueryInspector.tap_sql/2` on your query (with all the options passed) to see the exact search string and add an index to it. For example for nullable title, the GIN index could look like:\n\n    ```sql\n    CREATE INDEX index_gin_posts_title\n    ON posts USING GIN (to_tsvector('english', COALESCE(title, '')));\n    ```","ref":"Torus.html#full_text_dynamic/5-optimizations"},{"type":"macro","title":"Torus.ilike/5","doc":"Case-insensitive pattern matching search using\n[PostgreSQL `ILIKE`](https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-LIKE) operator.\n\n**Doesn't clean the term, so it needs to be sanitized before being passed in. See\n[LIKE-injections](https://githubengineering.com/like-injection/).**","ref":"Torus.html#ilike/5"},{"type":"macro","title":"Examples - Torus.ilike/5","doc":"iex> insert_posts!(titles: [\"Wand\", \"Magic wand\", \"Owl\"])\n    ...> Post\n    ...> |> Torus.ilike([p], [p.title], \"wan%\")\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Wand\"]\n\n    iex> insert_posts!([%{title: \"hogwarts\", body: nil}, %{title: nil, body: \"HOGWARTS\"}])\n    ...> Post\n    ...> |> Torus.ilike([p], [p.title, p.body], \"%OGWART%\")\n    ...> |> select([p], %{title: p.title, body: p.body})\n    ...> |> order_by(:id)\n    ...> |> Repo.all()\n    [%{title: \"hogwarts\", body: nil}, %{title: nil, body: \"HOGWARTS\"}]\n\n    iex> insert_post!(title: \"MaGiC\")\n    ...> Post\n    ...> |> Torus.ilike([p], p.title, \"magi%\")\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"MaGiC\"]","ref":"Torus.html#ilike/5-examples"},{"type":"macro","title":"Optimizations - Torus.ilike/5","doc":"See `like/5` optimization section for more details.","ref":"Torus.html#ilike/5-optimizations"},{"type":"macro","title":"Torus.like/5","doc":"Case-sensitive pattern matching search using [PostgreSQL `LIKE`](https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-LIKE) operator.\n\n**Doesn't clean the term, so it needs to be sanitized before being passed in, see\n[LIKE-injections](https://githubengineering.com/like-injection/)**","ref":"Torus.html#like/5"},{"type":"macro","title":"Examples - Torus.like/5","doc":"iex> insert_posts!([%{title: \"hogwarts\", body: nil}, %{title: nil, body: \"HOGWARTS\"}])\n    ...> Post\n    ...> |> Torus.like([p], [p.title, p.body], \"%OGWART%\")\n    ...> |> select([p], p.body)\n    ...> |> Repo.all()\n    [\"HOGWARTS\"]","ref":"Torus.html#like/5-examples"},{"type":"macro","title":"Optimizations - Torus.like/5","doc":"- `like/5` is case-sensitive, so it can take advantage of B-tree indexes when there\nis no wildcard (%) at the beginning of the search term, prefer it over `ilike/5` if\npossible.\n\n  Adding a B-tree index:\n\n    ```sql\n    CREATE INDEX index_posts_on_title ON posts (title);\n    ```\n\n- Use `GIN` or `GiST` Index with `pg_trgm` extension for LIKE and ILIKE.\n\n  When searching for substrings (%word%), B-tree indexes won't help. Instead,\n  use trigram indexing (`pg_trgm` extension):\n\n    ```sql\n    CREATE EXTENSION IF NOT EXISTS pg_trgm;\n    CREATE INDEX posts_title_trgm_idx ON posts USING GIN (title gin_trgm_ops);\n    ```\n\n- If using prefix search, convert data to lowercase and use B-tree index for\n  case-insensitive search:\n\n    ```sql\n    ALTER TABLE posts ADD COLUMN title_lower TEXT GENERATED ALWAYS AS (LOWER(title)) STORED;\n    CREATE INDEX index_posts_on_title ON posts (title_lower);\n    ```\n\n    ```elixir\n    Torus.like([p], [p.title_lower], \"hogwarts%\")\n    ```\n\n- Use full-text search for large text fields, see `full_text_dynamic/5` for more\ndetails.","ref":"Torus.html#like/5-optimizations"},{"type":"macro","title":"Torus.similar_to/5","doc":"Similar to `like/5`, except that it interprets the pattern using the SQL standard's\ndefinition of a regular expression. SQL regular expressions are a curious cross between\nLIKE notation and common (POSIX) regular expression notation. See\n[PostgreSQL `SIMILAR TO`](https://postgresql.org/docs/current/interactive/functions-matching.html?fts_query=ilike#FUNCTIONS-SIMILARTO-REGEXP)","ref":"Torus.html#similar_to/5"},{"type":"macro","title":"Examples - Torus.similar_to/5","doc":"iex> insert_post!(body: \"abc\")\n    ...> Post\n    ...> |> Torus.similar_to([p], [p.title, p.body], \"%(b|d)%\")\n    ...> |> select([p], p.body)\n    ...> |> Repo.all()\n    [\"abc\"]","ref":"Torus.html#similar_to/5-examples"},{"type":"macro","title":"Optimizations - Torus.similar_to/5","doc":"- If regex is needed, use POSIX regex with `~` or `~*` operators since they _may_\nleverage GIN or GiST indexes in some cases. These operators will be introduced later on.\n- Use `ilike/5` or `like/5` when possible, `similar_to/5` almost always does full table scans\n- Filter and limit the result set as much as possible before calling `similar_to/5`","ref":"Torus.html#similar_to/5-optimizations"},{"type":"macro","title":"Torus.similarity/5","doc":"Case-insensitive similarity search using [PostgreSQL similarity functions](https://postgresql.org/docs/current/interactive/pgtrgm.html#PGTRGM-FUNCS-OPS).\n\n**You need to have pg_trgm extension installed.**","ref":"Torus.html#similarity/5"},{"type":"macro","title":"Options - Torus.similarity/5","doc":"* `:type` - similarity type. Possible options are:\n    - `:full` (default) - uses `similarity` function.\n    - `:word` - uses `word_similarity` function. If you're dealing with sentences\n    and you don't want the length of the strings to affect the search result.\n    - `:strict` - uses `strict_word_similarity` function. Prioritizes full matches,\n    forces extent boundaries to match word boundaries. Since we don't have\n    cross-word trigrams, this function actually returns greatest similarity between\n    first string and any continuous extent of words of the second string.\n  * `:order` - describes the ordering of the results. Possible values are\n    - `:desc` (default) - orders the results by similarity rank in descending order.\n    - `:asc` - orders the results by similarity rank in ascending order.\n    - `:none` - doesn't apply ordering and returns\n  * `:limit` - limits the number of results returned (PostgreSQL `LIMIT`). By\n  default limit is not applied and the results are above\n  `pg_trgm.similarity_threshold`, which defaults to 0.3.\n  * `:pre_filter` - whether or not to pre-filter the results:\n    - `false` (default) - omits pre-filtering and returns all results.\n    - `true` -  before applying the order, pre filters (using boolean\n  operators which potentially use GIN indexes) the result set.","ref":"Torus.html#similarity/5-options"},{"type":"macro","title":"Examples - Torus.similarity/5","doc":"iex> insert_post!(title: \"Hogwarts Shocker\", body: \"A spell disrupts the Quidditch Cup.\")\n    ...> insert_post!(title: \"Diagon Bombshell\", body: \"Secrets uncovered in the heart of Hogwarts.\")\n    ...> insert_post!(title: \"Completely unrelated\", body: \"No magic here!\")\n    ...>  Post\n    ...> |> Torus.similarity([p], [p.title, p.body], \"Diagon Bombshell\", limit: 1)\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Diagon Bombshell\"]\n\n    iex> insert_posts!([\"Wand\", \"Owl\", \"What an amazing cloak\"])\n    ...> Post\n    ...> |> Torus.similarity([p], [p.title], \"what a cloak\", pre_filter: true)\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"What an amazing cloak\"]\n\n    iex> insert_posts!([\"Magic wand\", \"Wand\", \"Owlyth\"])\n    ...> Post\n    ...> |> Torus.similarity([p], [p.title], \"wand\", pre_filter: true)\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Wand\", \"Magic wand\"]","ref":"Torus.html#similarity/5-examples"},{"type":"macro","title":"Optimizations - Torus.similarity/5","doc":"- Use `pre_filter: true` to pre-filter the results before applying the order.\nThis would significantly reduce the number of rows to order. The pre-filtering\nphase uses different (boolean) similarity operators which more actively leverage\nGIN indexes.\n- Use `limit` to limit the number of results returned.\n- Use `order: `:none` argument if you don't care about the order of the results.\nThe query will return all results that are above the similarity threshold, which\nyou can set globally via `SET pg_trgm.similarity_threshold = 0.3;`.\n- When `order: `:desc` (default) and the limit is not set, the query will do a full\ntable scan, so it's recommended to manually limit the results (by applying `where`\nclauses to as little rows as possible).","ref":"Torus.html#similarity/5-optimizations"},{"type":"macro","title":"Adding an index - Torus.similarity/5","doc":"```sql\nCREATE EXTENSION IF NOT EXISTS pg_trgm;\n\nCREATE INDEX index_posts_on_title ON posts USING GIN (title gin_trgm_ops);\n```","ref":"Torus.html#similarity/5-adding-an-index"},{"type":"macro","title":"Already implemented optimizations - Torus.similarity/5","doc":"- Instead of scanning all rows and computing similarity for each, we first filter\nwith `%` (which could use a GIN index), and next we compute (potentially expensive)\nthe ranks using `similarity` function and order the results.","ref":"Torus.html#similarity/5-already-implemented-optimizations"},{"type":"macro","title":"Torus.substring/3","doc":"The substring function with three parameters provides extraction of a substring\nthat matches an SQL regular expression pattern. The function can be written\naccording to standard SQL syntax:\n\n```sql\nsubstring('foobar' similar '%#\"o_b#\"%' escape '#')   oob\nsubstring('foobar' similar '#\"o_b#\"%' escape '#')    NULL\n```","ref":"Torus.html#substring/3"},{"type":"macro","title":"Examples - Torus.substring/3","doc":"```elixir\ninsert_post!(title: \"Hello123World\")\nPost |> select([p], substring(p.title, \"[0-9]+\", \"#\")) |> Repo.all()\n[\"123\"]\n```","ref":"Torus.html#substring/3-examples"},{"type":"module","title":"Torus.QueryInspector","doc":"Helpers to debug your SQL queries. You can this module both while creating the\nqueries and directly in your running production shell once deployed, so that the\nexplain analyze returns more accurate results.","ref":"Torus.QueryInspector.html"},{"type":"function","title":"Torus.QueryInspector.substituted_sql/3","doc":"","ref":"Torus.QueryInspector.html#substituted_sql/3"},{"type":"function","title":"Torus.QueryInspector.tap_explain_analyze/3","doc":"Runs explain analyze on the query and prints it to the console. Returns the query.\n\n**Runs the query!**","ref":"Torus.QueryInspector.html#tap_explain_analyze/3"},{"type":"function","title":"Torus.QueryInspector.tap_sql/3","doc":"Converts the query to SQL and prints it to the console. Returns the query.","ref":"Torus.QueryInspector.html#tap_sql/3"},{"type":"function","title":"Torus.QueryInspector.tap_substituted_sql/3","doc":"Substitutes the parameters in the query and prints the SQL to the console. Returns the query.\nThe SQL is in its raw form and can be directly executed by postgres.","ref":"Torus.QueryInspector.html#tap_substituted_sql/3"},{"type":"extras","title":"Changelog","doc":"# v0.2.1\n\n`similarity/5` search is now fully tested and customizable. `full_text_dynamic/5` is up next.\n\n# Changelog for Torus v0.2.0\n\nTorus now supports full text search, ilike, and similarity search.","ref":"changelog.html"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.2"}}